`define OP  2:0 
`define PC 11:4
`define RD 14:12
`define RA 18:16
`define RB 22:20
`define IM 35:24
`define OPHALT 3
`define OPSW 40
`define OPLW 41
`define OPBNE 42
`define OPJALR 43

module risc_cpu 
#(parameter DATA_WIDTH=16, parameter ADDR_WIDTH=5)
(
	input clk, reset_n,
	output [2:0] rd,
	output [(DATA_WIDTH-1):0] qd,
	output rwe,
	output cpu_active,
	output [1:0] cpu_state,
	output reg [(DATA_WIDTH-1):0] ncycles
);

wire [(ADDR_WIDTH-1):0] addr;
wire [2:0] ra, rb;
wire  [(DATA_WIDTH-1):0] q, qi, qa, qb;
wire [47:0] IF;
reg [47:0] ID;
reg we;

assign cpu_active = active;
assign cpu_state = state;

single_port_ram dev1(
	.data(qb),
	.addr(addr),
	.we(we),
	.clk(clk),
	.q(q)
);

assign rwe = state==IFE;

regfile dev2(
	.clk(clk),
	.we(rwe),
	.ra(ra),
	.qa(qa),
	.rb(rb),
	.qb(qb),
	.rd(rd),
	.qd(qd)
);

reg [4:0] pc;
//wire [(DATA_WIDTH-1):0] q;
wire [2:0] opcode;

// Instruction Fetch (IF)

reg [15:0] inst;
reg active;
wire ma2 = (ID[`OPSW] || ID[`OPLW]);
//wire [15:0] qi;
assign qi = (active? (ma2? inst: q): 16'h0000);
assign opcode = qi[15:13];

// opcodes	
`include "opcodes.txt"

wire op_add = opcode==ADD;
wire op_nand = opcode==NAND;
wire op_sw = opcode==SW;
wire op_lw = opcode==LW;
wire op_bne = opcode==BNE;
wire op_jalr = opcode==JALR;

wire flag1 = op_sw||op_bne;   // no write-back
wire flag2 = op_add||op_nand; // RRR format
wire [2:0] rd1 = flag1? 3'h0: qi[12:10];
assign ra = qi[9:7];
assign rb = flag2? qi[2:0]: (flag1? qi[12:10]: 3'h0);

assign IF[`OP] = opcode;
assign IF[`OPHALT] = (qi==16'he071); // is halt
assign IF[`PC] = pc;
assign IF[`RD] = rd1;
assign IF[`RA] = ra;
assign IF[`RB] = rb;
assign IF[`IM] = qi[9:0];
assign IF[`OPSW] = op_sw;
assign IF[`OPLW] = op_lw;
assign IF[`OPBNE] = op_bne;
assign IF[`OPJALR] = op_jalr;


wire memory_access = (IF[`OPSW] || IF[`OPLW]);

//reg [(ADDR_WIDTH-1):0] jump;
wire [(ADDR_WIDTH-1):0] jump = do_jump? qa[(ADDR_WIDTH-1):0]: (ID[`OPBNE]? d1[(ADDR_WIDTH-1):0]: 5'h00);
wire do_branch = ID[`OPBNE] && (qa!=qb);
wire do_jump = ID[`OPJALR];


wire [4:0] nextpc = pc + 1'b1;
wire [4:0] newpc = (do_jump||do_branch? jump: nextpc);

reg [4:0] maddr;
assign addr = (state==MEM? d1: newpc );


wire [15:0] d1;

wire [4:0] npc = ID[`PC] + 1'b1;
assign rd = ID[`RD];

alu dev3(
  .a(qa),
  .b(qb),
  .pc({11'h000,npc}),
  .im10(ID[`IM]),
  .d(d1),
  .opcode(ID[`OP])
);

// writeback register multiplexer
assign qd = (ID[`OPLW]? q: d1);


reg [1:0] state;
localparam
	RESET = 2'h0,
	IFE = 2'h1,
	MEM = 2'h2,
	HALT = 2'h3;

wire [47:0] BLANK = 48'h0000_0000_0000;


always @(posedge clk, negedge reset_n)
begin
  if (!reset_n)
  begin
   active <= 1'b0;
   inst <= 16'h0000;
   pc <= 5'h1F;
   ID <= BLANK;
   state <= RESET;
   end
  else 
	case (state)
	RESET:
	begin
		active <= 1'b0;
		state <= IFE;
	end
	IFE:
	begin
		active <= 1'b1;
		pc <= addr;
		if (do_jump)
		begin
			if (ID[`OPHALT]) state <= HALT;
			ID <= BLANK;
		end
		else if (do_branch) ID <= BLANK;
		else 
		begin
			ID <= IF;
			if (memory_access) 
			begin 
				state <= MEM;
				we <= IF[`OPSW];
			end
		end
    end
    MEM:
    begin
		state <= IFE;
		inst <= q;
		we <= 1'b0;
	end
	default:
	begin
		active <= 1'b0;
		state <= HALT;
	end
  endcase
end

always @(posedge clk or negedge reset_n)
begin
	if (!reset_n) ncycles <= 16'h0000;
	else if (state==IFE || state==MEM) ncycles <= ncycles + 1'b1;
end

endmodule
